@@ -94,6 +94,18 @@ func configureXTransport(proxy *Proxy, config *Config) error {
 		if err != nil {
 			return fmt.Errorf("Unable to parse the HTTP proxy URL [%v]", config.HTTPProxyURL)
 		}
+
+		// Pre-resolve proxy hostname using bootstrap resolvers if it's a domain
+		if httpProxyURL.Hostname() != "" && ParseIP(httpProxyURL.Hostname()) == nil {
+			ips, ttl, err := proxy.xTransport.resolve(httpProxyURL.Hostname(), proxy.xTransport.useIPv4, proxy.xTransport.useIPv6)
+			if err != nil {
+				dlog.Warnf("Unable to resolve HTTP proxy hostname [%s] using bootstrap resolvers: %v", httpProxyURL.Hostname(), err)
+			} else if len(ips) > 0 {
+				proxy.xTransport.saveCachedIPs(httpProxyURL.Hostname(), ips, ttl)
+				dlog.Infof("Resolved HTTP proxy hostname [%s] to [%s] using bootstrap resolvers", httpProxyURL.Hostname(), ips[0])
+			}
+		}
+
 		proxy.xTransport.httpProxyFunction = http.ProxyURL(httpProxyURL)
 	}
 
@@ -159,6 +171,11 @@ func configureServerParams(proxy *Proxy, config *Config) {
 	proxy.blockedQueryResponse = config.BlockedQueryResponse
 	proxy.timeout = time.Duration(config.Timeout) * time.Millisecond
 	proxy.maxClients = config.MaxClients
+	proxy.timeoutLoadReduction = config.TimeoutLoadReduction
+	if proxy.timeoutLoadReduction < 0.0 || proxy.timeoutLoadReduction > 1.0 {
+		dlog.Warnf("timeout_load_reduction must be between 0.0 and 1.0, using default 0.75")
+		proxy.timeoutLoadReduction = 0.75
+	}
 	proxy.mainProto = "udp"
 	if config.ForceTCP {
 		proxy.mainProto = "tcp"
@@ -178,16 +195,21 @@ func configureLoadBalancing(proxy *Proxy, config *Config) {
 	lbStrategy := LBStrategy(DefaultLBStrategy)
 	switch lbStrategyStr := strings.ToLower(config.LBStrategy); lbStrategyStr {
 	case "":
-		// default
+		// default - WP2 is now the default strategy
+		dlog.Noticef("Using default Weighted Power of Two (WP2) load balancing strategy")
 	case "p2":
 		lbStrategy = LBStrategyP2{}
 	case "ph":
 		lbStrategy = LBStrategyPH{}
 	case "fastest":
+		// "fastest" kept for backward compatibility with older configs
+		fallthrough
 	case "first":
 		lbStrategy = LBStrategyFirst{}
 	case "random":
 		lbStrategy = LBStrategyRandom{}
+	case "wp2":
+		lbStrategy = LBStrategyWP2{}
 	default:
 		if strings.HasPrefix(lbStrategyStr, "p") {
 			n, err := strconv.ParseInt(strings.TrimPrefix(lbStrategyStr, "p"), 10, 32)
