@@ -7,6 +7,8 @@ import (
 	"html"
 	"net"
 	"net/http"
+	"os"
+	"runtime"
 	"sort"
 	"strings"
 	"sync"
@@ -17,25 +19,6 @@ import (
 	"github.com/miekg/dns"
 )
 
-// sanitizeString - Sanitizes user input to prevent XSS attacks
-func sanitizeString(input string) string {
-	// HTML escape to prevent XSS
-	escaped := html.EscapeString(input)
-	// Additional validation for domain names - only allow valid domain characters
-	if strings.Contains(input, ".") { // Likely a domain name
-		// Remove any non-domain characters
-		var result strings.Builder
-		for _, r := range escaped {
-			if (r >= 'a' && r <= 'z') || (r >= 'A' && r <= 'Z') ||
-				(r >= '0' && r <= '9') || r == '.' || r == '-' || r == '_' {
-				result.WriteRune(r)
-			}
-		}
-		return result.String()
-	}
-	return escaped
-}
-
 // MonitoringUIConfig - Configuration for the monitoring UI
 type MonitoringUIConfig struct {
 	Enabled            bool   `toml:"enabled"`
@@ -89,6 +72,9 @@ type MetricsCollector struct {
 
 	// Prometheus metrics (optional)
 	prometheusEnabled bool
+
+	// Runtime context
+	proxy *Proxy
 }
 
 // QueryLogEntry - Entry for the query log
@@ -114,6 +100,20 @@ func (q *QueryLogEntry) EstimateMemoryUsage() int64 {
 		len(q.Server))
 }
 
+type resolverSnapshot struct {
+	name          string
+	proto         string
+	total         uint64
+	failed        uint64
+	success       float64
+	avgObservedMs float64
+	lastUpdate    time.Time
+	lastAction    time.Time
+	status        string
+	score         float64
+	ageSeconds    float64
+}
+
 // MonitoringUI - Handles the monitoring UI
 type MonitoringUI struct {
 	config           MonitoringUIConfig
@@ -124,6 +124,9 @@ type MonitoringUI struct {
 	clientsMutex     sync.Mutex
 	proxy            *Proxy
 
+	// Mutex for all WebSocket write operations to prevent races
+	writesMutex sync.Mutex
+
 	// WebSocket broadcast rate limiting
 	broadcastMutex    sync.Mutex
 	lastBroadcast     time.Time
@@ -170,6 +173,7 @@ func NewMonitoringUI(proxy *Proxy) *MonitoringUI {
 		cachedMetrics: make(map[string]interface{}),
 		// Initialize Prometheus
 		prometheusEnabled: proxy.monitoringUI.PrometheusEnabled,
+		proxy:             proxy,
 	}
 
 	dlog.Debugf("Metrics collector initialized with privacy level: %d", metricsCollector.privacyLevel)
@@ -264,20 +268,17 @@ func (ui *MonitoringUI) Stop() error {
 }
 
 // UpdateMetrics - Updates metrics with a new query
-func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, start time.Time) {
+func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg) {
 	if !ui.config.Enabled {
 		return
 	}
 
-	dlog.Debugf("Updating metrics for query: %s", pluginsState.qName)
-
 	mc := ui.metricsCollector
 	now := time.Now()
 
 	// Update counters (total queries, cache, QPS) - separate lock
 	mc.countersMutex.Lock()
 	mc.totalQueries++
-	dlog.Debugf("Total queries now: %d", mc.totalQueries)
 
 	// Update queries per second
 	elapsed := now.Sub(mc.lastQueriesTime).Seconds()
@@ -292,16 +293,28 @@ func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, s
 			mc.lastQueriesCount = mc.totalQueries
 			mc.lastQueriesTime = now
 		}
-		dlog.Debugf("Updated QPS: %.2f", mc.queriesPerSecond)
 	}
 
 	// Update cache hits/misses
-	if pluginsState.cacheHit {
-		mc.cacheHits++
-		dlog.Debugf("Cache hit, total hits: %d", mc.cacheHits)
-	} else {
-		mc.cacheMisses++
-		dlog.Debugf("Cache miss, total misses: %d", mc.cacheMisses)
+	// Only count cache statistics for queries that participate in caching:
+	// - Cache hits (cacheHit == true)
+	// - Cache misses (queries that went to a DNS server: serverName != "-")
+	// This excludes blocked queries (REJECT/DROP) that never reach the cache or server
+	shouldCountCacheStats := pluginsState.cacheHit || pluginsState.serverName != "-"
+	if shouldCountCacheStats {
+		if pluginsState.cacheHit {
+			mc.cacheHits++
+		} else {
+			mc.cacheMisses++
+		}
+	}
+
+	// Update blocked queries count
+	// Only count truly blocked queries: REJECT (blocked by name/IP) and DROP (dropped)
+	// CLOAK is not counted as it redirects queries rather than blocking them
+	if pluginsState.returnCode == PluginsReturnCodeReject ||
+		pluginsState.returnCode == PluginsReturnCodeDrop {
+		mc.blockCount++
 	}
 	mc.countersMutex.Unlock()
 
@@ -317,18 +330,22 @@ func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, s
 		}
 		mc.queryTypesMutex.Lock()
 		mc.queryTypes[qType]++
-		dlog.Debugf("Query type %s, count: %d", qType, mc.queryTypes[qType])
 		mc.queryTypesMutex.Unlock()
 	} else {
 		dlog.Debugf("No question in message or message is nil")
 	}
 
 	// Update response time - back to counters lock
-	responseTime := time.Since(start).Milliseconds()
+	responseTime := time.Since(pluginsState.requestStart).Milliseconds()
+
+	// Cap at timeout to handle system sleep/suspend
+	maxResponseTime := pluginsState.timeout.Milliseconds()
+	if responseTime > maxResponseTime {
+		responseTime = maxResponseTime
+	}
 	mc.countersMutex.Lock()
 	mc.responseTimeSum += uint64(responseTime)
 	mc.responseTimeCount++
-	dlog.Debugf("Response time: %dms, avg: %.2fms", responseTime, float64(mc.responseTimeSum)/float64(mc.responseTimeCount))
 	mc.countersMutex.Unlock()
 
 	// Update server stats - separate lock
@@ -336,21 +353,15 @@ func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, s
 		mc.serverMutex.Lock()
 		mc.serverQueryCount[pluginsState.serverName]++
 		mc.serverResponseTime[pluginsState.serverName] += uint64(responseTime)
-		dlog.Debugf("Server %s, queries: %d, avg response: %.2fms",
-			pluginsState.serverName,
-			mc.serverQueryCount[pluginsState.serverName],
-			float64(mc.serverResponseTime[pluginsState.serverName])/float64(mc.serverQueryCount[pluginsState.serverName]))
 		mc.serverMutex.Unlock()
-	} else {
-		dlog.Debugf("No server name or server is '-'")
 	}
 
 	// Update top domains - separate lock
 	if mc.privacyLevel < 2 {
-		sanitizedDomain := sanitizeString(pluginsState.qName)
+		// Store domain name directly - no sanitization needed for internal metrics
+		domainName := pluginsState.qName
 		mc.domainMutex.Lock()
-		mc.topDomains[sanitizedDomain]++
-		dlog.Debugf("Domain %s, count: %d", sanitizedDomain, mc.topDomains[sanitizedDomain])
+		mc.topDomains[domainName]++
 		mc.domainMutex.Unlock()
 	}
 
@@ -397,13 +408,14 @@ func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, s
 		}
 
 		entry := QueryLogEntry{
-			Timestamp:    now,
-			ClientIP:     clientIP,
-			Domain:       sanitizeString(pluginsState.qName),
-			Type:         sanitizeString(qType),
-			ResponseCode: sanitizeString(returnCode),
+			Timestamp: now,
+			ClientIP:  clientIP,
+			// HTML escape only the fields that will be displayed in web UI
+			Domain:       html.EscapeString(pluginsState.qName),
+			Type:         qType,      // DNS types are safe, no escaping needed
+			ResponseCode: returnCode, // DNS response codes are safe, no escaping needed
 			ResponseTime: responseTime,
-			Server:       sanitizeString(pluginsState.serverName),
+			Server:       html.EscapeString(pluginsState.serverName),
 			CacheHit:     pluginsState.cacheHit,
 		}
 
@@ -430,8 +442,6 @@ func (ui *MonitoringUI) UpdateMetrics(pluginsState PluginsState, msg *dns.Msg, s
 			mc.currentMemoryBytes -= oldEntry.EstimateMemoryUsage()
 		}
 
-		dlog.Debugf("Added query log entry, total entries: %d, memory usage: %d bytes",
-			len(mc.recentQueries), mc.currentMemoryBytes)
 		mc.queryLogMutex.Unlock()
 	}
 
@@ -473,6 +483,10 @@ func (mc *MetricsCollector) generatePrometheusMetrics() string {
 	var result strings.Builder
 
 	// Write help and type information for each metric
+	result.WriteString("# HELP dnscrypt_proxy_build_info A metric with a constant '1' value labeled by version, goversion from which dnscrypt_proxy was built, and the goos and goarch for the build.\n")
+	result.WriteString("# TYPE dnscrypt_proxy_build_info gauge\n")
+	result.WriteString(fmt.Sprintf("dnscrypt_proxy_build_info{goarch=\"%s\" goos=\"%s\" goversion=\"%s\" version=\"%s\"} 1\n", runtime.GOARCH, runtime.GOOS, runtime.Version(), AppVersion))
+
 	result.WriteString("# HELP dnscrypt_proxy_queries_total Total number of DNS queries processed\n")
 	result.WriteString("# TYPE dnscrypt_proxy_queries_total counter\n")
 	result.WriteString(fmt.Sprintf("dnscrypt_proxy_queries_total %d\n", totalQueries))
@@ -510,8 +524,9 @@ func (mc *MetricsCollector) generatePrometheusMetrics() string {
 	result.WriteString("# HELP dnscrypt_proxy_server_queries_total Total queries per server\n")
 	result.WriteString("# TYPE dnscrypt_proxy_server_queries_total counter\n")
 	for server, count := range mc.serverQueryCount {
-		sanitizedServer := sanitizeString(server)
-		result.WriteString(fmt.Sprintf("dnscrypt_proxy_server_queries_total{server=\"%s\"} %d\n", sanitizedServer, count))
+		// For Prometheus labels, escape quotes and backslashes to prevent label injection
+		escapedServer := strings.ReplaceAll(strings.ReplaceAll(server, "\\", "\\\\"), "\"", "\\\"")
+		result.WriteString(fmt.Sprintf("dnscrypt_proxy_server_queries_total{server=\"%s\"} %d\n", escapedServer, count))
 	}
 
 	result.WriteString("# HELP dnscrypt_proxy_server_response_time_average_ms Average response time per server in milliseconds\n")
@@ -519,8 +534,9 @@ func (mc *MetricsCollector) generatePrometheusMetrics() string {
 	for server, count := range mc.serverQueryCount {
 		if count > 0 {
 			avgTime := float64(mc.serverResponseTime[server]) / float64(count)
-			sanitizedServer := sanitizeString(server)
-			result.WriteString(fmt.Sprintf("dnscrypt_proxy_server_response_time_average_ms{server=\"%s\"} %.2f\n", sanitizedServer, avgTime))
+			// For Prometheus labels, escape quotes and backslashes to prevent label injection
+			escapedServer := strings.ReplaceAll(strings.ReplaceAll(server, "\\", "\\\\"), "\"", "\\\"")
+			result.WriteString(fmt.Sprintf("dnscrypt_proxy_server_response_time_average_ms{server=\"%s\"} %.2f\n", escapedServer, avgTime))
 		}
 	}
 	mc.serverMutex.RUnlock()
@@ -530,8 +546,8 @@ func (mc *MetricsCollector) generatePrometheusMetrics() string {
 	result.WriteString("# HELP dnscrypt_proxy_query_type_total Total queries per DNS record type\n")
 	result.WriteString("# TYPE dnscrypt_proxy_query_type_total counter\n")
 	for qtype, count := range mc.queryTypes {
-		sanitizedQtype := sanitizeString(qtype)
-		result.WriteString(fmt.Sprintf("dnscrypt_proxy_query_type_total{type=\"%s\"} %d\n", sanitizedQtype, count))
+		// DNS query types are safe alphanumeric values, no escaping needed
+		result.WriteString(fmt.Sprintf("dnscrypt_proxy_query_type_total{type=\"%s\"} %d\n", qtype, count))
 	}
 	mc.queryTypesMutex.RUnlock()
 
@@ -552,6 +568,229 @@ func (mc *MetricsCollector) generatePrometheusMetrics() string {
 	return result.String()
 }
 
+func determineResolverStatus(total uint64, successRate float64, lastUpdate, lastAction, now time.Time) string {
+	staleThreshold := 5 * time.Minute
+	refTime := lastUpdate
+	if refTime.IsZero() {
+		refTime = lastAction
+	}
+
+	if total == 0 {
+		if refTime.IsZero() {
+			return "idle"
+		}
+		if !refTime.IsZero() && now.Sub(refTime) > staleThreshold {
+			return "stale"
+		}
+		return "warming"
+	}
+
+	status := "healthy"
+	switch {
+	case successRate >= 0.97:
+		status = "healthy"
+	case successRate >= 0.9:
+		status = "degraded"
+	default:
+		status = "failing"
+	}
+
+	if !refTime.IsZero() && now.Sub(refTime) > staleThreshold {
+		status = "stale"
+	}
+
+	return status
+}
+
+func resolverStatusRank(status string) int {
+	switch status {
+	case "failing":
+		return 0
+	case "degraded":
+		return 1
+	case "stale":
+		return 2
+	case "warming":
+		return 3
+	case "healthy":
+		return 4
+	case "idle":
+		return 5
+	default:
+		return 6
+	}
+}
+
+func (mc *MetricsCollector) collectResolverSnapshots() ([]resolverSnapshot, map[string]resolverSnapshot) {
+	snapshots := make([]resolverSnapshot, 0)
+	index := make(map[string]resolverSnapshot)
+
+	if mc.proxy == nil {
+		return snapshots, index
+	}
+
+	mc.proxy.serversInfo.RLock()
+	defer mc.proxy.serversInfo.RUnlock()
+
+	now := time.Now()
+	for _, server := range mc.proxy.serversInfo.inner {
+		if server == nil {
+			continue
+		}
+
+		total := server.totalQueries
+		failed := server.failedQueries
+		successRate := 1.0
+		if total > 0 {
+			successRate = float64(total-failed) / float64(total)
+		}
+
+		lastUpdate := server.lastUpdateTime
+		lastAction := server.lastActionTS
+		score := mc.proxy.serversInfo.calculateServerScore(server)
+		status := determineResolverStatus(total, successRate, lastUpdate, lastAction, now)
+		ageSeconds := -1.0
+		refTime := lastUpdate
+		if refTime.IsZero() {
+			refTime = lastAction
+		}
+		if !refTime.IsZero() {
+			ageSeconds = now.Sub(refTime).Seconds()
+		}
+
+		snapshot := resolverSnapshot{
+			name:       server.Name,
+			proto:      server.Proto.String(),
+			total:      total,
+			failed:     failed,
+			success:    successRate,
+			lastUpdate: lastUpdate,
+			lastAction: lastAction,
+			status:     status,
+			score:      score,
+			ageSeconds: ageSeconds,
+		}
+
+		snapshots = append(snapshots, snapshot)
+		index[server.Name] = snapshot
+	}
+
+	sort.Slice(snapshots, func(i, j int) bool {
+		if rankI, rankJ := resolverStatusRank(snapshots[i].status), resolverStatusRank(snapshots[j].status); rankI != rankJ {
+			return rankI < rankJ
+		}
+		if snapshots[i].score != snapshots[j].score {
+			return snapshots[i].score > snapshots[j].score
+		}
+		return snapshots[i].name < snapshots[j].name
+	})
+
+	return snapshots, index
+}
+
+func (mc *MetricsCollector) collectCacheStats(cacheHitRatio float64, cacheHits, cacheMisses uint64) map[string]interface{} {
+	stats := map[string]interface{}{
+		"enabled":         false,
+		"configured_size": 0,
+		"entries":         0,
+		"capacity":        0,
+		"cache_hit_ratio": cacheHitRatio,
+		"cache_hits":      cacheHits,
+		"cache_misses":    cacheMisses,
+	}
+
+	if mc.proxy == nil {
+		return stats
+	}
+
+	stats["enabled"] = mc.proxy.cache
+	stats["configured_size"] = mc.proxy.cacheSize
+	stats["max_ttl"] = mc.proxy.cacheMaxTTL
+	stats["min_ttl"] = mc.proxy.cacheMinTTL
+	stats["neg_max_ttl"] = mc.proxy.cacheNegMaxTTL
+	stats["neg_min_ttl"] = mc.proxy.cacheNegMinTTL
+
+	if cachedResponses.cache != nil {
+		stats["entries"] = cachedResponses.cache.Len()
+		stats["capacity"] = cachedResponses.cache.Capacity()
+	}
+
+	return stats
+}
+
+func (mc *MetricsCollector) collectSourceRefresh() []map[string]interface{} {
+	if mc.proxy == nil || len(mc.proxy.sources) == 0 {
+		return nil
+	}
+
+	results := make([]map[string]interface{}, 0, len(mc.proxy.sources))
+	now := time.Now()
+
+	for _, source := range mc.proxy.sources {
+		if source == nil {
+			continue
+		}
+
+		source.RLock()
+		name := source.name
+		cacheFile := source.cacheFile
+		nextRefresh := source.refresh
+		cacheTTL := source.cacheTTL
+		source.RUnlock()
+
+		var lastRefresh time.Time
+		var errorMessage string
+		if cacheFile != "" {
+			if fi, err := os.Stat(cacheFile); err == nil {
+				lastRefresh = fi.ModTime()
+			} else {
+				errorMessage = err.Error()
+			}
+		}
+
+		ageSeconds := -1.0
+		if !lastRefresh.IsZero() {
+			ageSeconds = now.Sub(lastRefresh).Seconds()
+		}
+
+		status := "ok"
+		switch {
+		case errorMessage != "":
+			status = "error"
+		case lastRefresh.IsZero():
+			status = "unknown"
+		case !nextRefresh.IsZero() && nextRefresh.Before(now):
+			status = "due"
+		case cacheTTL > 0 && lastRefresh.Add(cacheTTL).Before(now):
+			status = "stale"
+		}
+
+		entry := map[string]interface{}{
+			"name":        name,
+			"cache_file":  cacheFile,
+			"age_seconds": ageSeconds,
+			"status":      status,
+		}
+		if !lastRefresh.IsZero() {
+			entry["last_refresh"] = lastRefresh
+		}
+		if !nextRefresh.IsZero() {
+			entry["next_refresh"] = nextRefresh
+		}
+		if errorMessage != "" {
+			entry["error"] = errorMessage
+		}
+
+		results = append(results, entry)
+	}
+
+	sort.Slice(results, func(i, j int) bool {
+		return results[i]["name"].(string) < results[j]["name"].(string)
+	})
+
+	return results
+}
+
 // invalidateCache - Marks the cache as stale (call when data changes)
 func (mc *MetricsCollector) invalidateCache() {
 	mc.cacheMutex.Lock()
@@ -561,14 +800,11 @@ func (mc *MetricsCollector) invalidateCache() {
 
 // GetMetrics - Returns the current metrics
 func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
-	dlog.Debugf("GetMetrics called")
-
 	// Check cache first
 	mc.cacheMutex.RLock()
 	if time.Since(mc.cacheLastUpdate) < mc.cacheTTL && mc.cachedMetrics != nil {
 		cached := mc.cachedMetrics
 		mc.cacheMutex.RUnlock()
-		dlog.Debugf("Returning cached metrics")
 		return cached
 	}
 	mc.cacheMutex.RUnlock()
@@ -585,8 +821,6 @@ func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
 	startTime := mc.startTime
 	mc.countersMutex.RUnlock()
 
-	dlog.Debugf("GetMetrics - total queries: %d", totalQueries)
-
 	// Calculate average response time
 	var avgResponseTime float64
 	if responseTimeCount > 0 {
@@ -600,47 +834,27 @@ func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
 		cacheHitRatio = float64(cacheHits) / float64(totalCacheQueries)
 	}
 
-	// Calculate per-server metrics sorted by increasing average response time
-	serverMetrics := make([]map[string]interface{}, 0)
+	cacheStats := mc.collectCacheStats(cacheHitRatio, cacheHits, cacheMisses)
+	resolverSnapshots, resolverIndex := mc.collectResolverSnapshots()
 
-	// Create a slice of server performance data
-	type serverPerf struct {
-		name    string
-		queries uint64
-		avgTime float64
-	}
-
-	// Read server data with its own lock
+	// Update resolver snapshots with observed average response times.
 	mc.serverMutex.RLock()
-	serverPerfs := make([]serverPerf, 0, len(mc.serverQueryCount))
 	for server, count := range mc.serverQueryCount {
 		avgTime := float64(0)
 		if count > 0 {
 			avgTime = float64(mc.serverResponseTime[server]) / float64(count)
 		}
-		serverPerfs = append(serverPerfs, serverPerf{
-			name:    server,
-			queries: count,
-			avgTime: avgTime,
-		})
+		if snapshot, ok := resolverIndex[server]; ok {
+			snapshot.avgObservedMs = avgTime
+			resolverIndex[server] = snapshot
+		}
 	}
 	mc.serverMutex.RUnlock()
 
-	// Sort by increasing average response time (faster servers first)
-	sort.Slice(serverPerfs, func(i, j int) bool {
-		if serverPerfs[i].avgTime != serverPerfs[j].avgTime {
-			return serverPerfs[i].avgTime < serverPerfs[j].avgTime
+	for i, snapshot := range resolverSnapshots {
+		if updated, ok := resolverIndex[snapshot.name]; ok {
+			resolverSnapshots[i] = updated
 		}
-		return serverPerfs[i].name < serverPerfs[j].name
-	})
-
-	// Convert to map for JSON output
-	for _, sp := range serverPerfs {
-		serverMetrics = append(serverMetrics, map[string]interface{}{
-			"name":            sp.name,
-			"queries":         sp.queries,
-			"avg_response_ms": sp.avgTime,
-		})
 	}
 
 	// Get top domains (limited to 20) sorted by decreasing count
@@ -671,7 +885,7 @@ func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
 		count := 0
 		for _, dc := range domainCounts {
 			topDomainsList = append(topDomainsList, map[string]interface{}{
-				"domain": sanitizeString(dc.domain),
+				"domain": html.EscapeString(dc.domain),
 				"count":  dc.count,
 			})
 			count++
@@ -724,6 +938,35 @@ func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
 	copy(recentQueries, mc.recentQueries)
 	mc.queryLogMutex.RUnlock()
 
+	resolverHealth := make([]map[string]interface{}, 0, len(resolverSnapshots))
+	for _, snapshot := range resolverSnapshots {
+		entry := map[string]interface{}{
+			"name":           snapshot.name,
+			"proto":          snapshot.proto,
+			"status":         snapshot.status,
+			"success_rate":   snapshot.success,
+			"total_queries":  snapshot.total,
+			"failed_queries": snapshot.failed,
+			"score":          snapshot.score,
+		}
+		if snapshot.avgObservedMs > 0 {
+			entry["avg_response_ms"] = snapshot.avgObservedMs
+		}
+		if snapshot.ageSeconds >= 0 {
+			entry["age_seconds"] = snapshot.ageSeconds
+		}
+		if !snapshot.lastUpdate.IsZero() {
+			entry["last_update"] = snapshot.lastUpdate
+		}
+		if !snapshot.lastAction.IsZero() {
+			entry["last_action"] = snapshot.lastAction
+		}
+		resolverHealth = append(resolverHealth, entry)
+	}
+
+	sourceRefresh := mc.collectSourceRefresh()
+	generatedAt := time.Now().UTC()
+
 	// Return all metrics and cache the result
 	metrics := map[string]interface{}{
 		"total_queries":      totalQueries,
@@ -734,19 +977,21 @@ func (mc *MetricsCollector) GetMetrics() map[string]interface{} {
 		"cache_misses":       cacheMisses,
 		"avg_response_time":  avgResponseTime,
 		"blocked_queries":    blockCount,
-		"servers":            serverMetrics,
 		"top_domains":        topDomainsList,
 		"query_types":        queryTypesList,
 		"recent_queries":     recentQueries,
+		"cache_stats":        cacheStats,
+		"resolver_health":    resolverHealth,
+		"sources":            sourceRefresh,
+		"generated_at":       generatedAt,
 	}
 
 	// Cache the computed metrics
 	mc.cacheMutex.Lock()
 	mc.cachedMetrics = metrics
-	mc.cacheLastUpdate = time.Now()
+	mc.cacheLastUpdate = generatedAt
 	mc.cacheMutex.Unlock()
 
-	dlog.Debugf("Computed and cached new metrics")
 	return metrics
 }
 
@@ -772,8 +1017,6 @@ func setStaticCacheHeaders(w http.ResponseWriter, maxAge int) {
 
 // handleTestQuery - Handles test query requests for debugging
 func (ui *MonitoringUI) handleTestQuery(w http.ResponseWriter, r *http.Request) {
-	dlog.Debugf("Adding test query")
-
 	// Test queries modify state - no cache
 	setDynamicCacheHeaders(w)
 
@@ -782,16 +1025,18 @@ func (ui *MonitoringUI) handleTestQuery(w http.ResponseWriter, r *http.Request)
 	msg.SetQuestion("test.example.com.", dns.TypeA)
 
 	// Create a fake plugin state
+	testStart := time.Now().Add(-10 * time.Millisecond)
 	pluginsState := PluginsState{
-		qName:       "test.example.com",
-		serverName:  "cloudflare",
-		clientProto: "udp",
-		questionMsg: msg,
-		cacheHit:    false,
+		qName:        "test.example.com",
+		serverName:   "cloudflare",
+		clientProto:  "udp",
+		questionMsg:  msg,
+		cacheHit:     false,
+		requestStart: testStart,
 	}
 
 	// Update metrics
-	ui.UpdateMetrics(pluginsState, msg, time.Now().Add(-10*time.Millisecond))
+	ui.UpdateMetrics(pluginsState, msg)
 
 	// Return success
 	w.Header().Set("Content-Type", "text/plain")
@@ -800,8 +1045,6 @@ func (ui *MonitoringUI) handleTestQuery(w http.ResponseWriter, r *http.Request)
 
 // handleRoot - Handles the root path
 func (ui *MonitoringUI) handleRoot(w http.ResponseWriter, r *http.Request) {
-	dlog.Debugf("Received root request from %s", r.RemoteAddr)
-
 	// Set CORS headers
 	setCORSHeaders(w)
 
@@ -857,15 +1100,12 @@ func (ui *MonitoringUI) handleMetrics(w http.ResponseWriter, r *http.Request) {
 		return
 	}
 
-	dlog.Debugf("Sending metrics response (%d bytes)", len(jsonData))
-
 	// If it's a JSONP request, wrap the JSON in the callback function
 	if callback != "" {
 		w.Header().Set("Content-Type", "application/javascript")
 		w.Write([]byte(callback + "("))
 		w.Write(jsonData)
 		w.Write([]byte(");"))
-		dlog.Debugf("Sent JSONP response with callback: %s", callback)
 	} else {
 		// Regular JSON response
 		w.Write(jsonData)
@@ -883,32 +1123,24 @@ func (ui *MonitoringUI) handleWebSocket(w http.ResponseWriter, r *http.Request)
 		return
 	}
 
-	// Configure upgrader with more permissive settings
-	upgrader := websocket.Upgrader{
-		ReadBufferSize:  1024,
-		WriteBufferSize: 1024,
-		CheckOrigin: func(r *http.Request) bool {
-			return true // Allow all origins
-		},
-	}
-
-	conn, err := upgrader.Upgrade(w, r, nil)
+	conn, err := ui.upgrader.Upgrade(w, r, nil)
 	if err != nil {
 		dlog.Warnf("WebSocket upgrade error: %v", err)
 		return
 	}
 
-	// Set read/write deadlines
-	conn.SetReadDeadline(time.Now().Add(120 * time.Second))
-	conn.SetWriteDeadline(time.Now().Add(10 * time.Second))
-
+	// Register the client
 	ui.clientsMutex.Lock()
 	ui.clients[conn] = true
 	ui.clientsMutex.Unlock()
 
 	// Send initial metrics
-	metrics := ui.metricsCollector.GetMetrics()
-	if err := conn.WriteJSON(metrics); err != nil {
+	ui.writesMutex.Lock()
+	conn.SetWriteDeadline(time.Now().Add(10 * time.Second))
+	err = conn.WriteJSON(ui.metricsCollector.GetMetrics())
+	ui.writesMutex.Unlock()
+
+	if err != nil {
 		dlog.Warnf("WebSocket initial write error: %v", err)
 		conn.Close()
 		ui.clientsMutex.Lock()
@@ -927,78 +1159,43 @@ func (ui *MonitoringUI) handleWebSocket(w http.ResponseWriter, r *http.Request)
 			dlog.Debugf("WebSocket connection closed and cleaned up")
 		}()
 
-		// Create a ping handler to keep the connection alive
-		conn.SetPingHandler(func(data string) error {
-			dlog.Debugf("Received ping from client")
-			return conn.WriteControl(websocket.PongMessage, []byte{}, time.Now().Add(5*time.Second))
-		})
-
-		// Create a pong handler to respond to server pings
-		conn.SetPongHandler(func(data string) error {
+		// Set up ping/pong handlers for keep-alive (using WebSocket protocol level)
+		conn.SetReadDeadline(time.Now().Add(120 * time.Second))
+		conn.SetPongHandler(func(string) error {
 			dlog.Debugf("Received pong from client")
 			conn.SetReadDeadline(time.Now().Add(120 * time.Second))
 			return nil
 		})
 
 		for {
-			// Reset read deadline for each message
-			conn.SetReadDeadline(time.Now().Add(120 * time.Second))
-
-			// Read message
-			messageType, message, err := conn.ReadMessage()
+			// Read message from client
+			var msg map[string]interface{}
+			err := conn.ReadJSON(&msg)
 			if err != nil {
 				if websocket.IsUnexpectedCloseError(err, websocket.CloseGoingAway, websocket.CloseAbnormalClosure) {
 					dlog.Warnf("WebSocket unexpected close error: %v", err)
-				} else {
-					dlog.Debugf("WebSocket read error (normal): %v", err)
 				}
 				break
 			}
 
-			// Handle ping message from client
-			if messageType == websocket.TextMessage {
-				var msg map[string]interface{}
-				if err := json.Unmarshal(message, &msg); err == nil {
-					if msgType, ok := msg["type"].(string); ok && msgType == "ping" {
-						dlog.Debugf("Received ping message from client")
-						// Send a pong response
-						conn.SetWriteDeadline(time.Now().Add(5 * time.Second))
-						if err := conn.WriteJSON(map[string]string{"type": "pong"}); err != nil {
-							dlog.Warnf("Error sending pong: %v", err)
-						}
+			// Handle ping message from client (application level)
+			if msgType, ok := msg["type"].(string); ok && msgType == "ping" {
+				dlog.Debugf("Received ping message from client")
 
-						// Also send updated metrics
-						conn.SetWriteDeadline(time.Now().Add(5 * time.Second))
-						if err := conn.WriteJSON(ui.metricsCollector.GetMetrics()); err != nil {
-							dlog.Warnf("Error sending metrics after ping: %v", err)
-						}
-					}
-				}
-			}
-		}
-	}()
-
-	// Send periodic pings to keep the connection alive
-	go func() {
-		ticker := time.NewTicker(30 * time.Second)
-		defer ticker.Stop()
-
-		for {
-			select {
-			case <-ticker.C:
-				ui.clientsMutex.Lock()
-				if _, exists := ui.clients[conn]; !exists {
-					ui.clientsMutex.Unlock()
-					return // Connection is closed, stop the goroutine
-				}
-				ui.clientsMutex.Unlock()
-
-				// Send ping
+				// Send pong response and updated metrics
+				ui.writesMutex.Lock()
 				conn.SetWriteDeadline(time.Now().Add(5 * time.Second))
-				if err := conn.WriteControl(websocket.PingMessage, []byte{}, time.Now().Add(5*time.Second)); err != nil {
-					dlog.Debugf("Error sending ping: %v", err)
-					return
+				if err := conn.WriteJSON(map[string]string{"type": "pong"}); err != nil {
+					ui.writesMutex.Unlock()
+					dlog.Warnf("Error sending pong: %v", err)
+					break
 				}
+
+				conn.SetWriteDeadline(time.Now().Add(5 * time.Second))
+				if err := conn.WriteJSON(ui.metricsCollector.GetMetrics()); err != nil {
+					dlog.Warnf("Error sending metrics after ping: %v", err)
+				}
+				ui.writesMutex.Unlock()
 			}
 		}
 	}()
@@ -1039,8 +1236,6 @@ func (ui *MonitoringUI) handlePrometheus(w http.ResponseWriter, r *http.Request)
 	// Write metrics
 	w.WriteHeader(http.StatusOK)
 	w.Write([]byte(metrics))
-
-	dlog.Debugf("Served Prometheus metrics (%d bytes)", len(metrics))
 }
 
 // basicAuthMiddleware - Adds basic authentication to the HTTP server
@@ -1103,10 +1298,14 @@ func (ui *MonitoringUI) scheduleBroadcast() {
 func (ui *MonitoringUI) broadcastMetrics() {
 	metrics := ui.metricsCollector.GetMetrics()
 
+	ui.writesMutex.Lock()
+	defer ui.writesMutex.Unlock()
+
 	ui.clientsMutex.Lock()
 	defer ui.clientsMutex.Unlock()
 
 	for client := range ui.clients {
+		client.SetWriteDeadline(time.Now().Add(5 * time.Second))
 		err := client.WriteJSON(metrics)
 		if err != nil {
 			dlog.Debugf("WebSocket write error: %v", err)
