@@ -1,6 +1,8 @@
 package main
 
 import (
+	"errors"
+	"fmt"
 	"math/rand"
 	"net"
 	"strings"
@@ -13,13 +15,14 @@ import (
 )
 
 type CloakedName struct {
-	target     string
-	ipv4       []net.IP
-	ipv6       []net.IP
-	lastUpdate *time.Time
-	lineNo     int
-	isIP       bool
-	PTR        []string
+	target      string
+	ipv4        []net.IP
+	ipv6        []net.IP
+	lastUpdate4 *time.Time
+	lastUpdate6 *time.Time
+	lineNo      int
+	isIP        bool
+	PTR         []string
 }
 
 type PluginCloak struct {
@@ -27,6 +30,11 @@ type PluginCloak struct {
 	patternMatcher *PatternMatcher
 	ttl            uint32
 	createPTR      bool
+
+	// Hot-reloading support
+	configFile     string
+	configWatcher  *ConfigWatcher
+	stagingMatcher *PatternMatcher
 }
 
 func (plugin *PluginCloak) Name() string {
@@ -38,20 +46,35 @@ func (plugin *PluginCloak) Description() string {
 }
 
 func (plugin *PluginCloak) Init(proxy *Proxy) error {
-	dlog.Noticef("Loading the set of cloaking rules from [%s]", proxy.cloakFile)
-	bin, err := ReadTextFile(proxy.cloakFile)
+	plugin.configFile = proxy.cloakFile
+	dlog.Noticef("Loading the set of cloaking rules from [%s]", plugin.configFile)
+
+	lines, err := ReadTextFile(plugin.configFile)
 	if err != nil {
 		return err
 	}
+
 	plugin.ttl = proxy.cloakTTL
 	plugin.createPTR = proxy.cloakedPTR
 	plugin.patternMatcher = NewPatternMatcher()
+
+	if err := plugin.loadRules(lines, plugin.patternMatcher); err != nil {
+		return err
+	}
+
+	return nil
+}
+
+// loadRules parses cloaking rules from text and adds them to a pattern matcher
+func (plugin *PluginCloak) loadRules(lines string, patternMatcher *PatternMatcher) error {
 	cloakedNames := make(map[string]*CloakedName)
-	for lineNo, line := range strings.Split(string(bin), "\n") {
+
+	for lineNo, line := range strings.Split(lines, "\n") {
 		line = TrimAndStripInlineComments(line)
 		if len(line) == 0 {
 			continue
 		}
+
 		var target string
 		parts := strings.FieldsFunc(line, unicode.IsSpace)
 		if len(parts) == 2 {
@@ -65,20 +88,19 @@ func (plugin *PluginCloak) Init(proxy *Proxy) error {
 			dlog.Errorf("Syntax error in cloaking rules at line %d -- Missing name or target", 1+lineNo)
 			continue
 		}
+
 		line = strings.ToLower(line)
 		cloakedName, found := cloakedNames[line]
 		if !found {
 			cloakedName = &CloakedName{}
 		}
+
 		ip := net.ParseIP(target)
 		if ip != nil {
 			if ipv4 := ip.To4(); ipv4 != nil {
-				cloakedName.ipv4 = append((*cloakedName).ipv4, ipv4)
-			} else if ipv6 := ip.To16(); ipv6 != nil {
-				cloakedName.ipv6 = append((*cloakedName).ipv6, ipv6)
+				cloakedName.ipv4 = append(cloakedName.ipv4, ipv4)
 			} else {
-				dlog.Errorf("Invalid IP address in cloaking rule at line %d", 1+lineNo)
-				continue
+				cloakedName.ipv6 = append(cloakedName.ipv6, ip)
 			}
 			cloakedName.isIP = true
 		} else {
@@ -96,7 +118,7 @@ func (plugin *PluginCloak) Init(proxy *Proxy) error {
 			reversed, _ := dns.ReverseAddr(ip.To4().String())
 			ptrLine = strings.TrimSuffix(reversed, ".")
 		} else {
-			reversed, _ := dns.ReverseAddr(cloakedName.ipv6[0].To16().String())
+			reversed, _ := dns.ReverseAddr(cloakedName.ipv6[0].String())
 			ptrLine = strings.TrimSuffix(reversed, ".")
 		}
 		ptrQueryLine := ptrEntryToQuery(ptrLine)
@@ -109,11 +131,13 @@ func (plugin *PluginCloak) Init(proxy *Proxy) error {
 		ptrCloakedName.lineNo = lineNo + 1
 		cloakedNames[ptrQueryLine] = ptrCloakedName
 	}
+
 	for line, cloakedName := range cloakedNames {
-		if err := plugin.patternMatcher.Add(line, cloakedName, cloakedName.lineNo); err != nil {
+		if err := patternMatcher.Add(line, cloakedName, cloakedName.lineNo); err != nil {
 			return err
 		}
 	}
+
 	return nil
 }
 
@@ -127,19 +151,84 @@ func ptrNameToFQDN(ptrLine string) string {
 }
 
 func (plugin *PluginCloak) Drop() error {
+	if plugin.configWatcher != nil {
+		plugin.configWatcher.RemoveFile(plugin.configFile)
+	}
 	return nil
 }
 
-func (plugin *PluginCloak) Reload() error {
+// PrepareReload loads new cloaking rules into staging matcher but doesn't apply them yet
+func (plugin *PluginCloak) PrepareReload() error {
+	// Read the configuration file
+	lines, err := SafeReadTextFile(plugin.configFile)
+	if err != nil {
+		return fmt.Errorf("error reading config file during reload preparation: %w", err)
+	}
+
+	// Create new staging pattern matcher
+	plugin.stagingMatcher = NewPatternMatcher()
+
+	// Load rules into staging matcher
+	if err := plugin.loadRules(lines, plugin.stagingMatcher); err != nil {
+		return fmt.Errorf("error parsing config during reload preparation: %w", err)
+	}
+
 	return nil
 }
 
+// ApplyReload atomically replaces the active pattern matcher with the staging one
+func (plugin *PluginCloak) ApplyReload() error {
+	if plugin.stagingMatcher == nil {
+		return errors.New("no staged configuration to apply")
+	}
+
+	// Use write lock to swap pattern matchers
+	plugin.Lock()
+	plugin.patternMatcher = plugin.stagingMatcher
+	plugin.stagingMatcher = nil
+	plugin.Unlock()
+
+	dlog.Noticef("Applied new configuration for plugin [%s]", plugin.Name())
+	return nil
+}
+
+// CancelReload cleans up any staging resources
+func (plugin *PluginCloak) CancelReload() {
+	plugin.stagingMatcher = nil
+}
+
+// Reload implements hot-reloading for the plugin
+func (plugin *PluginCloak) Reload() error {
+	dlog.Noticef("Reloading configuration for plugin [%s]", plugin.Name())
+
+	// Prepare the new configuration
+	if err := plugin.PrepareReload(); err != nil {
+		plugin.CancelReload()
+		return err
+	}
+
+	// Apply the new configuration
+	return plugin.ApplyReload()
+}
+
+// GetConfigPath returns the path to the plugin's configuration file
+func (plugin *PluginCloak) GetConfigPath() string {
+	return plugin.configFile
+}
+
+// SetConfigWatcher sets the config watcher for this plugin
+func (plugin *PluginCloak) SetConfigWatcher(watcher *ConfigWatcher) {
+	plugin.configWatcher = watcher
+}
+
 func (plugin *PluginCloak) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
 	question := msg.Question[0]
 	if question.Qclass != dns.ClassINET || question.Qtype == dns.TypeNS || question.Qtype == dns.TypeSOA {
 		return nil
 	}
 	now := time.Now()
+
+	// Use read lock for thread-safe access to patternMatcher
 	plugin.RLock()
 	_, _, xcloakedName := plugin.patternMatcher.Eval(pluginsState.qName)
 	if xcloakedName == nil {
@@ -154,42 +243,60 @@ func (plugin *PluginCloak) Eval(pluginsState *PluginsState, msg *dns.Msg) error
 	}
 	cloakedName := xcloakedName.(*CloakedName)
 	ttl, expired := plugin.ttl, false
-	if cloakedName.lastUpdate != nil {
-		if elapsed := uint32(now.Sub(*cloakedName.lastUpdate).Seconds()); elapsed < ttl {
+	var lastUpdate *time.Time
+	switch question.Qtype {
+	case dns.TypeA:
+		lastUpdate = cloakedName.lastUpdate4
+	case dns.TypeAAAA:
+		lastUpdate = cloakedName.lastUpdate6
+	}
+	if lastUpdate != nil {
+		if elapsed := uint32(now.Sub(*lastUpdate).Seconds()); elapsed < ttl {
 			ttl -= elapsed
 		} else {
 			expired = true
 		}
 	}
-	if !cloakedName.isIP && ((cloakedName.ipv4 == nil && cloakedName.ipv6 == nil) || expired) {
+	synth := EmptyResponseFromMessage(msg)
+	if !cloakedName.isIP && ((question.Qtype == dns.TypeA && cloakedName.ipv4 == nil) ||
+		(question.Qtype == dns.TypeAAAA && cloakedName.ipv6 == nil) || expired) {
 		target := cloakedName.target
 		plugin.RUnlock()
-		foundIPs, err := net.LookupIP(target)
+		returnIPv4 := question.Qtype == dns.TypeA
+		returnIPv6 := question.Qtype == dns.TypeAAAA
+		foundIPs, _, err := pluginsState.xTransport.resolveUsingServers(
+			pluginsState.xTransport.mainProto,
+			target,
+			pluginsState.xTransport.internalResolvers,
+			returnIPv4,
+			returnIPv6,
+		)
 		if err != nil {
+			synth.Rcode = dns.RcodeServerFailure
+			pluginsState.synthResponse = synth
+			pluginsState.action = PluginsActionSynth
+			pluginsState.returnCode = PluginsReturnCodeCloak
 			return nil
 		}
+
+		// Use write lock to update cloakedName
 		plugin.Lock()
-		cloakedName.lastUpdate = &now
-		cloakedName.ipv4 = nil
-		cloakedName.ipv6 = nil
-		for _, foundIP := range foundIPs {
-			if ipv4 := foundIP.To4(); ipv4 != nil {
-				cloakedName.ipv4 = append(cloakedName.ipv4, foundIP)
-				if len(cloakedName.ipv4) >= 16 {
-					break
-				}
-			} else {
-				cloakedName.ipv6 = append(cloakedName.ipv6, foundIP)
-				if len(cloakedName.ipv6) >= 16 {
-					break
-				}
+		if len(foundIPs) > 0 {
+			n := Min(16, len(foundIPs))
+			switch question.Qtype {
+			case dns.TypeA:
+				cloakedName.lastUpdate4 = &now
+				cloakedName.ipv4 = foundIPs[:n]
+			case dns.TypeAAAA:
+				cloakedName.lastUpdate6 = &now
+				cloakedName.ipv6 = foundIPs[:n]
 			}
 		}
 		plugin.Unlock()
+
+		// Reacquire read lock
 		plugin.RLock()
 	}
-	plugin.RUnlock()
-	synth := EmptyResponseFromMessage(msg)
 	synth.Answer = []dns.RR{}
 	if question.Qtype == dns.TypeA {
 		for _, ip := range cloakedName.ipv4 {
@@ -213,6 +320,8 @@ func (plugin *PluginCloak) Eval(pluginsState *PluginsState, msg *dns.Msg) error
 			synth.Answer = append(synth.Answer, rr)
 		}
 	}
+	plugin.RUnlock()
+
 	rand.Shuffle(
 		len(synth.Answer),
 		func(i, j int) { synth.Answer[i], synth.Answer[j] = synth.Answer[j], synth.Answer[i] },
