@@ -11,8 +11,10 @@ import (
 	"github.com/miekg/dns"
 )
 
-const myResolverHost string = "resolver.dnscrypt.info."
-const nonexistentName string = "nonexistent-zone.dnscrypt-test."
+const (
+	myResolverHost  string = "resolver.dnscrypt.info."
+	nonexistentName string = "nonexistent-zone.dnscrypt-test."
+)
 
 func resolveQuery(server string, qName string, qType uint16, sendClientSubnet bool) (*dns.Msg, error) {
 	client := new(dns.Client)
@@ -52,8 +54,8 @@ func resolveQuery(server string, qName string, qType uint16, sendClientSubnet bo
 	options.SetUDPSize(uint16(MaxDNSPacketSize))
 
 	msg.Question[0] = dns.Question{Name: qName, Qtype: qType, Qclass: dns.ClassINET}
-	msg.Id = dns.Id()
 	for i := 0; i < 3; i++ {
+		msg.Id = dns.Id()
 		response, rtt, err := client.Exchange(msg, server)
 		if neterr, ok := err.(net.Error); ok && neterr.Timeout() {
 			client.ReadTimeout *= 2
@@ -139,6 +141,7 @@ func Resolve(server string, name string, singleResolver bool) {
 			fmt.Printf("Lying         : ")
 			response, err := resolveQuery(server, nonexistentName, dns.TypeA, false)
 			if err != nil {
+				fmt.Printf("[%v]", err)
 				break
 			}
 			if response.Rcode == dns.RcodeSuccess {
@@ -285,7 +288,7 @@ cname:
 		} else if len(mxs) > 1 {
 			fmt.Printf("%d mail servers found\n", len(mxs))
 		} else {
-			fmt.Println("1 mail servers found")
+			fmt.Println("1 mail server found")
 		}
 	}
 
