@@ -64,6 +64,9 @@ function handleError(error) {
     }
 }
 
+// Cache for the last non-empty recent queries
+let lastRecentQueries = [];
+
 // Safe update function that handles missing data
 function safeUpdateDashboard(data) {
     try {
@@ -72,6 +75,11 @@ function safeUpdateDashboard(data) {
             return;
         }
 
+        if (data.type === 'pong') {
+            console.log('Received pong message');
+            return;
+        }
+
         console.log('Updating dashboard with data');
 
         // Store the current scroll position before updates
@@ -88,34 +96,95 @@ function safeUpdateDashboard(data) {
 
         // Update overview stats with null checks
         const totalQueries = data.total_queries !== undefined ? data.total_queries : 0;
+        const blockedQueries = data.blocked_queries !== undefined ? data.blocked_queries : 0;
         const qps = data.queries_per_second !== undefined ? data.queries_per_second : 0;
         const uptime = data.uptime_seconds !== undefined ? data.uptime_seconds : 0;
         const avgResponseTime = data.avg_response_time !== undefined ? data.avg_response_time : 0;
 
-        document.getElementById('total-queries').textContent = totalQueries.toLocaleString();
-        document.getElementById('qps').textContent = qps.toFixed(2);
-        document.getElementById('uptime').textContent = formatUptime(uptime);
-        document.getElementById('avg-response-time').textContent = avgResponseTime.toFixed(2) + ' ms';
+        updateElementText('total-queries', formatNumber(totalQueries));
+        updateElementText('blocked-queries', formatNumber(blockedQueries));
+        updateElementText('qps', qps.toFixed(2));
+        updateElementText('uptime', formatUptime(uptime));
+        updateElementText('avg-response-time', formatMilliseconds(avgResponseTime));
+
+        const generatedAt = data.generated_at ? new Date(data.generated_at) : null;
+        const lastUpdatedEl = document.getElementById('last-updated');
+        if (lastUpdatedEl) {
+            if (generatedAt && !isNaN(generatedAt.getTime())) {
+                lastUpdatedEl.textContent = generatedAt.toLocaleString();
+            } else {
+                lastUpdatedEl.textContent = '-';
+            }
+        }
 
         // Update cache stats with null checks
         const cacheHitRatio = data.cache_hit_ratio !== undefined ? data.cache_hit_ratio : 0;
         const cacheHits = data.cache_hits !== undefined ? data.cache_hits : 0;
         const cacheMisses = data.cache_misses !== undefined ? data.cache_misses : 0;
+        const cacheStats = data.cache_stats || {};
 
-        document.getElementById('cache-hit-ratio').textContent = (cacheHitRatio * 100).toFixed(2) + '%';
-        document.getElementById('cache-hits').textContent = cacheHits.toLocaleString();
-        document.getElementById('cache-misses').textContent = cacheMisses.toLocaleString();
+        updateElementText('cache-hit-ratio', formatPercent(cacheHitRatio));
+        updateElementText('cache-hits', formatNumber(cacheHits));
+        updateElementText('cache-misses', formatNumber(cacheMisses));
+        updateElementText('cache-enabled', formatBoolean(cacheStats.enabled));
+        updateElementText('cache-configured-size', formatNumber(cacheStats.configured_size));
+        updateElementText('cache-entries', formatNumber(cacheStats.entries));
+        updateElementText('cache-capacity', formatNumber(cacheStats.capacity));
+        updateElementText('cache-ttl-range', formatTTLRange(cacheStats));
 
-        // Update server table
-        const serverTable = document.getElementById('server-table').getElementsByTagName('tbody')[0];
-        serverTable.innerHTML = '';
-        if (data.servers && Array.isArray(data.servers)) {
-            data.servers.forEach(server => {
-                const row = serverTable.insertRow();
-                row.insertCell(0).textContent = server.name || 'Unknown';
-                row.insertCell(1).textContent = (server.queries || 0).toLocaleString();
-                row.insertCell(2).textContent = (server.avg_response_ms || 0).toFixed(2) + ' ms';
+        // Update resolver health table
+        const resolverTable = document.getElementById('resolver-table').getElementsByTagName('tbody')[0];
+        resolverTable.innerHTML = '';
+        const resolverRows = Array.isArray(data.resolver_health) ? data.resolver_health : [];
+
+        if (resolverRows.length > 0) {
+            // Sort by total queries (desc), then avg response, name, and last seen.
+            const sortedResolvers = resolverRows.slice().sort((a, b) => {
+                const totalQueries = resolver => {
+                    if (typeof resolver.total_queries === 'number') {
+                        return resolver.total_queries;
+                    }
+                    if (typeof resolver.queries === 'number') {
+                        return resolver.queries;
+                    }
+                    return -1;
+                };
+                const avgResponse = resolver => {
+                    return typeof resolver.avg_response_ms === 'number' ? resolver.avg_response_ms : Number.POSITIVE_INFINITY;
+                };
+                const name = resolver => (resolver.name || '').toLowerCase();
+                const lastSeen = resolver => {
+                    const value = resolver.last_update;
+                    if (!value) {
+                        return 0;
+                    }
+                    const parsed = Date.parse(value);
+                    return Number.isNaN(parsed) ? 0 : parsed;
+                };
+
+                return (
+                    (totalQueries(b) - totalQueries(a)) ||
+                    (avgResponse(a) - avgResponse(b)) ||
+                    (name(a) > name(b) ? 1 : name(a) < name(b) ? -1 : 0) ||
+                    (lastSeen(b) - lastSeen(a))
+                );
             });
+
+            sortedResolvers.forEach(resolver => {
+                const row = resolverTable.insertRow();
+                row.insertCell(0).textContent = resolver.name || 'Unknown';
+                row.insertCell(1).textContent = formatStatus(resolver.status);
+                row.insertCell(2).textContent = formatPercent(resolver.success_rate);
+                row.insertCell(3).textContent = formatNumber(resolver.total_queries !== undefined ? resolver.total_queries : resolver.queries);
+                row.insertCell(4).textContent = formatNumber(resolver.failed_queries);
+                row.insertCell(5).textContent = formatMilliseconds(resolver.avg_response_ms);
+                row.insertCell(6).textContent = formatTimestamp(resolver.last_update);
+            });
+        } else {
+            const row = resolverTable.insertRow();
+            const cell = row.insertCell(0);
+            cell.colSpan = 7;
+            cell.textContent = 'No resolver data yet';
         }
 
         // Update query types table
@@ -140,11 +209,35 @@ function safeUpdateDashboard(data) {
             });
         }
 
+        // Update sources table
+        const sourcesTable = document.getElementById('sources-table').getElementsByTagName('tbody')[0];
+        sourcesTable.innerHTML = '';
+        if (data.sources && Array.isArray(data.sources) && data.sources.length > 0) {
+            data.sources.forEach(source => {
+                const row = sourcesTable.insertRow();
+                row.insertCell(0).textContent = source.name || '-';
+                row.insertCell(1).textContent = formatTimestamp(source.last_refresh);
+                row.insertCell(2).textContent = formatTimestamp(source.next_refresh);
+                row.insertCell(3).textContent = formatSourceStatus(source.status, source.error);
+                row.insertCell(4).textContent = formatAge(source.age_seconds);
+            });
+        } else {
+            const row = sourcesTable.insertRow();
+            const cell = row.insertCell(0);
+            cell.colSpan = 5;
+            cell.textContent = 'No source activity recorded yet';
+        }
+
         // Update recent queries table
         const queriesTable = document.getElementById('queries-table').getElementsByTagName('tbody')[0];
+        let queriesToShow = lastRecentQueries;
+        if (data.recent_queries && Array.isArray(data.recent_queries) && data.recent_queries.length > 0) {
+            lastRecentQueries = data.recent_queries;
+            queriesToShow = lastRecentQueries;
+        }
         queriesTable.innerHTML = '';
-        if (data.recent_queries && Array.isArray(data.recent_queries)) {
-            data.recent_queries.slice().reverse().forEach(query => {
+        if (queriesToShow && Array.isArray(queriesToShow)) {
+            queriesToShow.slice().reverse().forEach(query => {
                 const row = queriesTable.insertRow();
                 row.insertCell(0).textContent = query.timestamp ? new Date(query.timestamp).toLocaleTimeString() : '-';
                 row.insertCell(1).textContent = query.domain || '-';
@@ -152,7 +245,7 @@ function safeUpdateDashboard(data) {
                 row.insertCell(3).textContent = query.client_ip || '-';
                 row.insertCell(4).textContent = query.server || '-';
                 row.insertCell(5).textContent = query.response_code || '-';
-                row.insertCell(6).textContent = (query.response_time || 0) + ' ms';
+                row.insertCell(6).textContent = formatMilliseconds(query.response_time);
             });
         }
 
@@ -182,6 +275,120 @@ function formatUptime(seconds) {
     }
 }
 
+function updateElementText(id, value) {
+    const el = document.getElementById(id);
+    if (!el) {
+        return;
+    }
+    el.textContent = value !== undefined && value !== null && value !== '' ? value : '-';
+}
+
+function formatNumber(value) {
+    if (value === undefined || value === null) {
+        return '-';
+    }
+    const num = Number(value);
+    if (Number.isNaN(num)) {
+        return '-';
+    }
+    return num.toLocaleString();
+}
+
+function formatPercent(value) {
+    if (value === undefined || value === null) {
+        return '-';
+    }
+    const num = Number(value);
+    if (Number.isNaN(num)) {
+        return '-';
+    }
+    return (num * 100).toFixed(1) + '%';
+}
+
+function formatMilliseconds(value) {
+    if (value === undefined || value === null) {
+        return '-';
+    }
+    const num = Number(value);
+    if (Number.isNaN(num)) {
+        return '-';
+    }
+    return Math.round(num) + ' ms';
+}
+
+function formatBoolean(value) {
+    if (value === undefined || value === null) {
+        return '-';
+    }
+    return value ? 'Yes' : 'No';
+}
+
+function formatStatus(status) {
+    if (!status || typeof status !== 'string') {
+        return 'Unknown';
+    }
+    const normalized = status.replace(/_/g, ' ').toLowerCase();
+    return normalized.charAt(0).toUpperCase() + normalized.slice(1);
+}
+
+function formatTimestamp(value) {
+    if (!value) {
+        return '-';
+    }
+    try {
+        const date = new Date(value);
+        if (Number.isNaN(date.getTime())) {
+            return '-';
+        }
+        return date.toLocaleString();
+    } catch (error) {
+        return '-';
+    }
+}
+
+function formatAge(seconds) {
+    if (seconds === undefined || seconds === null) {
+        return '-';
+    }
+    const num = Number(seconds);
+    if (Number.isNaN(num) || num < 0) {
+        return '-';
+    }
+    const totalSeconds = Math.round(num);
+    const days = Math.floor(totalSeconds / 86400);
+    const hours = Math.floor((totalSeconds % 86400) / 3600);
+    const minutes = Math.floor((totalSeconds % 3600) / 60);
+    const secs = totalSeconds % 60;
+    const parts = [];
+    if (days > 0) parts.push(days + 'd');
+    if (hours > 0) parts.push(hours + 'h');
+    if (minutes > 0) parts.push(minutes + 'm');
+    parts.push(secs + 's');
+    return parts.join(' ');
+}
+
+function formatTTLRange(cacheStats) {
+    if (!cacheStats || typeof cacheStats !== 'object') {
+        return '-';
+    }
+    const parts = [];
+    if (cacheStats.min_ttl !== undefined && cacheStats.max_ttl !== undefined) {
+        parts.push('pos ' + cacheStats.min_ttl + 's-' + cacheStats.max_ttl + 's');
+    }
+    if (cacheStats.neg_min_ttl !== undefined && cacheStats.neg_max_ttl !== undefined) {
+        parts.push('neg ' + cacheStats.neg_min_ttl + 's-' + cacheStats.neg_max_ttl + 's');
+    }
+    return parts.length ? parts.join(' / ') : '-';
+}
+
+function formatSourceStatus(status, error) {
+    let label = formatStatus(status);
+    if (error) {
+        label += ' (' + error + ')';
+    }
+    return label;
+}
+
 // Simple direct data loading approach
 function loadData() {
     console.log('Loading data using simple approach');
@@ -273,23 +480,23 @@ function connectWebSocket() {
         console.log('WebSocket URL:', wsUrl);
 
         // Create WebSocket connection
-        var ws = new WebSocket(wsUrl);
+        var newWs = new WebSocket(wsUrl);
 
         // Connection opened
-        ws.onopen = function() {
+        newWs.onopen = function() {
             console.log('WebSocket connected successfully');
             wsReconnectAttempts = 0; // Reset reconnect attempts on successful connection
 
             // Send a ping to verify connection
             try {
-                ws.send(JSON.stringify({type: 'ping'}));
+                newWs.send(JSON.stringify({type: 'ping'}));
             } catch (e) {
                 console.error('Error sending ping:', e);
             }
         };
 
         // Listen for messages
-        ws.onmessage = function(event) {
+        newWs.onmessage = function(event) {
             try {
                 if (!event) {
                     console.warn('Received invalid WebSocket event');
@@ -310,12 +517,12 @@ function connectWebSocket() {
         };
 
         // Handle errors
-        ws.onerror = function(error) {
+        newWs.onerror = function(error) {
             console.error('WebSocket error occurred:', error);
         };
 
         // Connection closed
-        ws.onclose = function(event) {
+        newWs.onclose = function(event) {
             console.log('WebSocket disconnected, code:', event.code, 'reason:', event.reason || 'No reason provided');
 
             // Try to reconnect with exponential backoff
@@ -325,10 +532,8 @@ function connectWebSocket() {
                 console.log('Attempting to reconnect in ' + delay + 'ms (attempt ' + wsReconnectAttempts + '/' + maxReconnectAttempts + ')');
 
                 setTimeout(function() {
-                    var newWs = connectWebSocket();
-                    if (newWs) {
-                        // We can't update the global ws variable from here
-                        // Instead, we'll rely on the polling fallback
+                    ws = connectWebSocket();
+                    if (ws) {
                         console.log('New WebSocket connection established');
                     }
                 }, delay);
@@ -337,7 +542,7 @@ function connectWebSocket() {
             }
         };
 
-        return ws;
+        return newWs;
     } catch (error) {
         console.error('Failed to create WebSocket connection:', error);
         return null;
@@ -385,13 +590,20 @@ window.handlePollData = function(data) {
 
 // Initialize dashboard with default values
 function initializeDashboard() {
-    document.getElementById('total-queries').textContent = '0';
-    document.getElementById('qps').textContent = '0.00';
-    document.getElementById('uptime').textContent = '0s';
-    document.getElementById('avg-response-time').textContent = '0.00 ms';
-    document.getElementById('cache-hit-ratio').textContent = '0.00%';
-    document.getElementById('cache-hits').textContent = '0';
-    document.getElementById('cache-misses').textContent = '0';
+    updateElementText('total-queries', '0');
+    updateElementText('blocked-queries', '0');
+    updateElementText('qps', '0.00');
+    updateElementText('uptime', '0s');
+    updateElementText('avg-response-time', '0.00 ms');
+    updateElementText('cache-hit-ratio', '0.0%');
+    updateElementText('cache-hits', '0');
+    updateElementText('cache-misses', '0');
+    updateElementText('cache-enabled', '-');
+    updateElementText('cache-configured-size', '-');
+    updateElementText('cache-entries', '-');
+    updateElementText('cache-capacity', '-');
+    updateElementText('cache-ttl-range', '-');
+    updateElementText('last-updated', '-');
 }
 
 // Initialize with default values
@@ -446,4 +658,4 @@ setTimeout(function() {
             loadingIndicator.style.display = 'none';
         }, 5000);
     }
-}, 20000);
\ No newline at end of file
+}, 20000);
